﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Drawing;
using HeuristicLab.Persistence.Data;

namespace HeuristicLab.Persistence {
  internal abstract class ValueTransformer<TSource, TTarget> : TransformerBase {
    public override bool CanTransformType(Type type) {
      return typeof(TSource) == type;
    }
    public override PersistenceData ToData(object o, PersistenceMapper mapper) {
      return GetTarget((TSource)o, mapper);
    }

    public override object ToObject(PersistenceData data, PersistenceMapper mapper) {
      return GetSource((ValueData<TTarget>)data, mapper);
    }

    protected abstract ValueData<TTarget> GetTarget(TSource source, PersistenceMapper mapper);
    protected abstract TSource GetSource(ValueData<TTarget> target, PersistenceMapper mapper);
  }

  internal abstract class ValueTypeTransformer<T> : ValueTransformer<T, T> {
    protected override ValueData<T> GetTarget(T source, PersistenceMapper mapper) {
      var data = new ValueData<T>(Id);
      data.TypeId = mapper.GetTypeId(source.GetType());
      data.Value = source;
      return data;
    }
    protected override T GetSource(ValueData<T> target, PersistenceMapper mapper) {
      return target.Value;
    }
  }

  #region Primitive Value Types
  [Transformer("9FA1C6A8-517E-4623-AC1B-7E9AEF6ED13D", 200)]
  internal sealed class BoolTransformer : ValueTypeTransformer<bool> { }

  [Transformer("059633E9-12CB-43EC-8544-57746536E281", 201)]
  internal sealed class ByteTransformer : ValueTypeTransformer<byte> { }

  [Transformer("5DC2F3AC-0852-4B57-A861-D29CC814A94C", 202)]
  internal sealed class SByteTransformer : ValueTypeTransformer<sbyte> { }

  [Transformer("B862E642-A94A-4870-8065-06126A35A9E1", 203)]
  internal sealed class ShortTransformer : ValueTypeTransformer<short> { }

  [Transformer("D1D3062D-F1BB-4189-AE50-D6986E1DEB4E", 204)]
  internal sealed class UShortTransformer : ValueTypeTransformer<ushort> { }

  [Transformer("6C444645-3062-4D15-AD01-E6E1B0692A2B", 205)]
  internal sealed class CharTransformer : ValueTypeTransformer<char> { }

  [Transformer("649E73B2-EFA6-4E01-BCB4-4B29D652C9CB", 206)]
  internal sealed class IntTransformer : ValueTypeTransformer<int> { }

  [Transformer("BCF25010-81A2-49BC-88CC-407D3F393D5B", 207)]
  internal sealed class UIntTransformer : ValueTypeTransformer<uint> { }

  [Transformer("B6F6ACAE-755C-47EE-B8BF-7CDECBE19C30", 208)]
  internal sealed class LongTransformer : ValueTypeTransformer<long> { }

  [Transformer("82333ACA-F041-44E0-B365-27C399594BA7", 209)]
  internal sealed class ULongTransformer : ValueTypeTransformer<ulong> { }

  [Transformer("8FE91ECF-2261-4934-BECD-C38923B7A703", 210)]
  internal sealed class FloatTransformer : ValueTypeTransformer<float> { }

  [Transformer("070D23EA-7F38-46B7-A667-219BEF914E49", 211)]
  internal sealed class DoubleTransformer : ValueTypeTransformer<double> { }

  [Transformer("80368686-0531-4D52-B550-E772B8842E6D", 212)]
  internal sealed class DecimalTransformer : ValueTypeTransformer<decimal> { }

  [Transformer("BA55C7A6-C91E-4351-A889-E4A7E647F16D", 213)]
  internal sealed class DateTimeTransformer : ValueTypeTransformer<DateTime> { }

  [Transformer("0C91441B-2D97-432B-B493-D6EC483FC5AD", 214)]
  internal sealed class TimeSpanTransformer : ValueTypeTransformer<TimeSpan> { }
  #endregion

  #region String
  [Transformer("E75A594C-0034-4DAB-B28E-8F84F9F6DE8D", 215)]
  internal sealed class StringTransformer : ValueTransformer<string, uint> {
    protected override ValueData<uint> GetTarget(string source, PersistenceMapper mapper) {
      var data = new ValueData<uint>(Id);
      mapper.Cache(source, data);
      data.TypeId = mapper.GetTypeId(source.GetType());
      data.Value = mapper.GetStringId(source);
      return data;
    }
    protected override string GetSource(ValueData<uint> target, PersistenceMapper mapper) {
      var source = mapper.GetString(target.Value);
      mapper.Cache(target, source);
      return source;
    }
  }
  #endregion

  #region Enum
  [Transformer("93FF076B-BC4B-4C39-8C40-15E004468C98", 216)]
  internal sealed class EnumTransformer : ValueTransformer<object, uint> {
    public override bool CanTransformType(Type type) {
      return typeof(Enum).IsAssignableFrom(type);
    }
    protected override ValueData<uint> GetTarget(object source, PersistenceMapper mapper) {
      var type = source.GetType();
      var data = new ValueData<uint>(Id);
      data.TypeId = mapper.GetTypeId(type);
      data.Value = mapper.GetStringId(Enum.Format(type, source, "G"));
      return data;
    }
    protected override object GetSource(ValueData<uint> target, PersistenceMapper mapper) {
      return Enum.Parse(mapper.GetType(target.TypeId), mapper.GetString(target.Value));
    }
  }
  #endregion

  #region Color
  [Transformer("0B540EAC-79AB-40CF-8277-D2C81135FEB6", 217)]
  internal sealed class ColorTransformer : ValueTransformer<Color, int> {
    protected override ValueData<int> GetTarget(Color source, PersistenceMapper mapper) {
      var data = new ValueData<int>(Id);
      data.TypeId = mapper.GetTypeId(source.GetType());
      data.Value = source.ToArgb();
      return data;
    }
    protected override Color GetSource(ValueData<int> target, PersistenceMapper mapper) {
      return Color.FromArgb(target.Value);
    }
  }
  #endregion

  #region Type
  [Transformer("8D17FD28-383B-44E9-9BBF-B19D351C5E38", 218)]
  internal sealed class TypeTransformer : ValueTransformer<Type, ushort> {
    public override bool CanTransformType(Type type) {
      return typeof(Type).IsAssignableFrom(type);
    }
    protected override ValueData<ushort> GetTarget(Type source, PersistenceMapper mapper) {
      var data = new ValueData<ushort>(Id);
      mapper.Cache(source, data);
      data.Value = mapper.GetTypeId(source);
      return data;
    }
    protected override Type GetSource(ValueData<ushort> target, PersistenceMapper mapper) {
      var type = mapper.GetType(target.Value);
      mapper.Cache(target, type);
      return type;
    }
  }
  #endregion
}
